using System;
using System.Configuration;
using System.Net;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Channels;
using System.Threading;

using KnownObjects;
using Belikov.GenuineChannels;
using Belikov.GenuineChannels.BroadcastEngine;
using Belikov.GenuineChannels.DotNetRemotingLayer;
using Belikov.GenuineChannels.Logbook;

using Belikov.GenuineChannels.TransportContext;
using Belikov.GenuineChannels.Security;
using Belikov.GenuineChannels.Security.SSPI;

namespace Client
{
	/// <summary>
	/// ChatClient demostrates simple client application.
	/// </summary>
	class ChatClient : MarshalByRefObject, IMessageReceiver
	{
		/// <summary>
		/// The only instance.
		/// </summary>
		public static ChatClient Instance = new ChatClient();

		/// <summary>
		/// User name.
		/// </summary>
		public static string UserName;

		/// <summary>
		/// Password.
		/// </summary>
		public static string Password;

		/// <summary>
		/// Password.
		/// </summary>
		public static string Domain;

		/// <summary>
		/// Chat room.
		/// </summary>
		public static IChatRoom IChatRoom;

		/// <summary>
		/// A proxy to server business object.
		/// </summary>
		public static IChatServer IChatServer;

		/// <summary>
		/// To provide thread-safe access to ChatClient.IChatServer member.
		/// </summary>
		public static object IChatServerLock = new object();

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			// wait for the server
			Console.WriteLine("Sleep for 3 seconds.");
			Thread.Sleep(TimeSpan.FromSeconds(3));

			// setup .NET Remoting
			Console.WriteLine("Configuring Remoting environment...");
			System.Configuration.ConfigurationSettings.GetConfig("DNS");
			GenuineGlobalEventProvider.GenuineChannelsGlobalEvent += new GenuineChannelsGlobalEventHandler(GenuineChannelsEventHandler);
			GenuineGlobalLoggerContainer.Logger = new BinaryLog(@"c:\tmp\client.log", false, true);
			RemotingConfiguration.Configure("Client.exe.config");

			Console.WriteLine(".NET Remoting has been configured from Client.exe.config file.");

			// bind client's receiver
			RemotingServices.Marshal(ChatClient.Instance, "MessageReceiver.rem");

			ChatClient.IChatServer = (IChatServer) Activator.GetObject(typeof(IChatRoom),
				ConfigurationSettings.AppSettings["RemoteHostUri"] + "/ChatServer.rem");

			KeyProvider_SspiClient keyProvider_SspiClient = new KeyProvider_SspiClient(SspiFeatureFlags.Impersonation | SspiFeatureFlags.Delegation, SupportedSspiPackages.Kerberos, 
				null, ChatClient.IChatServer.GetAuthority(), null);
			ITransportContextProvider iTransportContextProvider = (ITransportContextProvider) ChannelServices.GetChannel("gtcp");
			iTransportContextProvider.ITransportContext.IKeyStore.SetKey("/Chat/Sspi", keyProvider_SspiClient);
			iTransportContextProvider.ITransportContext.SecuritySessionParameters = new SecuritySessionParameters("/Chat/Sspi");

			for(;;)
			{
				try
				{
					for ( ; ; )
					{
						try
						{
							Console.WriteLine("Please enter a user name. Press ENTER to use process's security context.");
							ChatClient.UserName = Console.ReadLine();
							if (ChatClient.UserName.Length > 0)
							{
								Console.WriteLine("Please enter the password:");
								ChatClient.Password = Console.ReadLine();
								Console.WriteLine("Please enter the name of the domain (or just an empty line):");
								ChatClient.Domain = Console.ReadLine();
							}

							if (ChatClient.UserName.Length <= 0)
								keyProvider_SspiClient.AuthIdentity = null;
							else
								keyProvider_SspiClient.AuthIdentity = new NetworkCredential(ChatClient.UserName, ChatClient.Password, ChatClient.Domain);

							// subscribe to the chat event
							lock(ChatClient.IChatServerLock)
							{
								ChatClient.IChatServer = (IChatServer) Activator.GetObject(typeof(IChatRoom),
									ConfigurationSettings.AppSettings["RemoteHostUri"] + "/ChatServer.rem");
								ChatClient.IChatRoom = ChatClient.IChatServer.EnterToChatRoom();
							}

							break;
						}
						catch (Exception ex)
						{
							Console.WriteLine("Error: {0}.", ex.Message);

							// some messages are outputted in event threads, do not confuse users
							Thread.Sleep(2000);
						}
					}

					for(;;)
					{
						Console.WriteLine("Enter a message to send or an empty string to exit.");

						string str = Console.ReadLine();
						if (str.Length <= 0)
							return ;

						string user = ChatClient.IChatRoom.SendMessage(str);
						Console.WriteLine("Account = {0}.", user);
					}
				}
				catch(Exception ex)
				{
					Console.WriteLine("\r\n\r\n---Exception: {0}. Stack trace: {1}.", ex.Message, ex.StackTrace);
				}

				Console.WriteLine("Next attempt to connect to the server will be in 3 seconds.");
				Thread.Sleep(3000);
			}
		}

		public static void GenuineChannelsEventHandler(object sender, GenuineEventArgs e)
		{
			if (e.SourceException == null)
				Console.WriteLine("\r\n\r\n---Global event: {0}\r\nRemote host: {1}", 
					e.EventType,
					e.HostInformation == null ? "<unknown>" : e.HostInformation.ToString());
			else
				Console.WriteLine("\r\n\r\n---Global event: {0}\r\nRemote host: {1}\r\nException: {2}", 
					e.EventType, 
					e.HostInformation == null ? "<unknown>" : e.HostInformation.ToString(), 
					e.SourceException);

			if (e.EventType == GenuineEventType.GeneralServerRestartDetected)
			{
				// server has been restarted so it does not know that we have been subscribed to
				// messages and ours nickname
				lock(ChatClient.IChatServerLock)
				{
					ChatClient.IChatServer = (IChatServer) Activator.GetObject(typeof(IChatRoom),
						ConfigurationSettings.AppSettings["RemoteHostUri"] + "/ChatServer.rem");
					ChatClient.IChatRoom = ChatClient.IChatServer.EnterToChatRoom();
				}
			}
		}

		/// <summary>
		/// Message receiver.
		/// It receives messages async and writes them separately from the main thread.
		/// But it does not matter for console application.
		/// </summary>
		/// <param name="message">The message.</param>
		public object ReceiveMessage(string message, string nickname)
		{
			Console.WriteLine("Message \"{0}\" from \"{1}\".", message, nickname);
			return null;
		}

		/// <summary>
		/// This is to insure that when created as a Singleton, the first instance never dies,
		/// regardless of the expired time.
		/// </summary>
		/// <returns></returns>
		public override object InitializeLifetimeService()
		{
			return null;
		}
	}
}
