using System;
using System.Configuration;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Messaging;
using System.Runtime.Remoting.Channels;
using System.Security.Principal;
using System.Threading;

using KnownObjects;
using Belikov.GenuineChannels;
using Belikov.GenuineChannels.BroadcastEngine;
using Belikov.GenuineChannels.Security;
using Belikov.GenuineChannels.Security.SSPI;
using Belikov.GenuineChannels.TransportContext;

namespace Server
{
	/// <summary>
	/// Represents a chat room.
	/// </summary>
	public class ChatRoom : MarshalByRefObject, IChatRoom
	{
		/// <summary>
		/// Constructs ChatRoom instance.
		/// </summary>
		public ChatRoom()
		{
			// bind server's methods
			this._dispatcher.BroadcastCallFinishedHandler += new BroadcastCallFinishedHandler(this.BroadcastCallFinishedHandler);
			this._dispatcher.CallIsAsync = true;

			if (this.IThirdParty == null)
				this.IThirdParty = (IThirdParty) Activator.GetObject(typeof(IThirdParty), ConfigurationSettings.AppSettings["SecondServerUri"] + "/TrackingServer.rem");
		}

		public IThirdParty IThirdParty;

		/// <summary>
		/// Chat members.
		/// </summary>
		private Dispatcher _dispatcher = new Dispatcher(typeof(IMessageReceiver));

		/// <summary>
		/// Attaches the client.
		/// </summary>
		/// <param name="nickname">Nickname.</param>
		public void AttachClient(string nickname)
		{
			string receiverUri = GenuineUtility.FetchCurrentRemoteUri() + "/MessageReceiver.rem";
			IMessageReceiver iMessageReceiver = (IMessageReceiver) Activator.GetObject(typeof(IMessageReceiver), receiverUri);
			this._dispatcher.Add((MarshalByRefObject) iMessageReceiver);

			GenuineUtility.CurrentSession["Nickname"] = nickname;
			Console.WriteLine("Client with nickname \"{0}\" has been registered.", nickname);
		}

		/// <summary>
		/// Sends message to all clients.
		/// </summary>
		/// <param name="message">Message to send.</param>
		/// <returns>Number of clients having received this message.</returns>
		public string SendMessage(string message)
		{
			// fetch the nickname
			string nickname = GenuineUtility.CurrentSession["Nickname"] as string;
			Console.WriteLine("Message \"{0}\" will be sent to all clients from {1}.", message, nickname);

			// get the established SecuritySession_SspiServer
			ITransportContext iTransportContext = GenuineUtility.CurrentRemoteHost.ITransportContext;
			SecuritySession_SspiServer currentServerSspiSession = (SecuritySession_SspiServer) GenuineUtility.CurrentRemoteHost.GetSecuritySession(GenuineUtility.CurrentInvocationSecuritySessionParameters.Name, null);

			// here is a very good place to check a principal
			WindowsPrincipal windowsPrincipal = new WindowsPrincipal(currentServerSspiSession.WindowsIdentity);
			if ( windowsPrincipal.IsInRole(@"YOURDOMAIN\Domain Admins") )
				Console.WriteLine("The user is a Domain Admin! Wow!");
			else
				Console.WriteLine("The user is not a Domain Admin.");

			// get the name of identity
			string identityName = WindowsIdentity.GetCurrent().Name;
			string keyName = "/Chat/SSPIUserKeys/" + GenuineUtility.CurrentRemoteUri;

			// ensure that the security environment is correct
			string establishedSession = GenuineUtility.CurrentSession["AuthorizedIdentity"] as string;
			if (establishedSession == null)
			{
				// if there is no security context, create it
				iTransportContext.IKeyStore.SetKey(keyName, 
					new KeyProvider_SspiClient(SspiFeatureFlags.Impersonation | SspiFeatureFlags.Delegation | SspiFeatureFlags.Encryption, SupportedSspiPackages.Kerberos,
					null, this.IThirdParty.GetTarget(), currentServerSspiSession));
				this.IThirdParty.CreateSecurityContext(keyName);
			}
			else if (establishedSession != null && establishedSession != identityName)
			{
				// if security session has been established but it uses another credential,
				// release and re-establish it
				GenuineUtility.CurrentRemoteHost.DestroySecuritySession(keyName);
				this.IThirdParty.ReleaseSecurityContext(keyName);
			}

			GenuineUtility.CurrentSession["AuthorizedIdentity"] = identityName;

			// use the setup security context
			using (new SecurityContextKeeper(new SecuritySessionParameters(keyName)))
			{
				this.IThirdParty.MessageHasBeenSent(message);
			}

			IMessageReceiver iMessageReceiver = (IMessageReceiver) this._dispatcher.TransparentProxy;
			iMessageReceiver.ReceiveMessage(message, nickname);

			return WindowsIdentity.GetCurrent().Name;
		}

		/// <summary>
		/// Called by broadcast dispatcher when all calls are performed.
		/// Does not undertake any actions.
		/// </summary>
		/// <param name="dispatcher">Source dipatcher.</param>
		/// <param name="message">Source message.</param>
		/// <param name="resultCollector">Call results.</param>
		public void BroadcastCallFinishedHandler(Dispatcher dispatcher, IMessage message, ResultCollector resultCollector)
		{
		}

	}
}
