using System;
using System.Collections;
using System.Security.Principal;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Channels;
using System.Runtime.Remoting.Lifetime;
using System.Runtime.Remoting.Messaging;

using KnownObjects;
using Belikov.GenuineChannels;
using Belikov.GenuineChannels.BroadcastEngine;
using Belikov.GenuineChannels.DotNetRemotingLayer;
using Belikov.GenuineChannels.Logbook;

using Belikov.GenuineChannels.TransportContext;
using Belikov.GenuineChannels.Security;
using Belikov.GenuineChannels.Security.SSPI;

namespace Server
{
	/// <summary>
	/// Chat server implements server that configures Genuine Server TCP Channel and implements
	/// chat server behavior.
	/// </summary>
	class TrackingServer : MarshalByRefObject, IThirdParty
	{
		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			try
			{
				// setup .NET remoting
				System.Configuration.ConfigurationSettings.GetConfig("DNS");
				GenuineGlobalEventProvider.GenuineChannelsGlobalEvent += new GenuineChannelsGlobalEventHandler(GenuineChannelsEventHandler);
				GenuineGlobalLoggerContainer.Logger = new BinaryLog(@"c:\tmp\trackingServer.log", false);
				RemotingConfiguration.Configure("SecondServer.exe.config");

				// bind the server
				RemotingServices.Marshal(new TrackingServer(), "TrackingServer.rem");

				Console.WriteLine("Server has been started. Press enter to exit.");
				Console.ReadLine();
			}
			catch(Exception ex)
			{
				Console.WriteLine("Exception: {0}. Stack trace: {1}.", ex.Message, ex.StackTrace);
			}
		}

		/// <summary>
		/// Catches Genuine Channels events and removes client session when
		/// user disconnects.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		public static void GenuineChannelsEventHandler(object sender, GenuineEventArgs e)
		{
			if (e.SourceException == null)
				Console.WriteLine("\r\n\r\n---Global event: {0}\r\nRemote host: {1}", 
					e.EventType,
					e.HostInformation == null ? "<unknown>" : e.HostInformation.ToString());
			else
				Console.WriteLine("\r\n\r\n---Global event: {0}\r\nRemote host: {1}\r\nException: {2}", 
					e.EventType, 
					e.HostInformation == null ? "<unknown>" : e.HostInformation.ToString(), 
					e.SourceException);

			if (e.EventType == GenuineEventType.GeneralConnectionClosed)
			{
				// the client disconnected
				string nickname = e.HostInformation["Nickname"] as string;
				if (nickname != null)
					Console.WriteLine("Client \"{0}\" has been disconnected.", nickname);
			}
		}

		/// <summary>
		/// Creates SSPI Server provider with the specified name.
		/// </summary>
		/// <param name="securityProviderName">The name of the Security Provider.</param>
		public void CreateSecurityContext(string securityProviderName)
		{
			ITransportContext iTransportContext = GenuineUtility.CurrentRemoteHost.ITransportContext;
			iTransportContext.IKeyStore.SetKey(securityProviderName, 
				new KeyProvider_SspiServer(SspiFeatureFlags.Impersonation | SspiFeatureFlags.Delegation | SspiFeatureFlags.Encryption, SupportedSspiPackages.Kerberos));
		}

		/// <summary>
		/// Releases the established Security Sessions with the specified name.
		/// </summary>
		/// <param name="securityProviderName">The name of the Security Provider.</param>
		public void ResetSecurityContext(string securityProviderName)
		{
			GenuineUtility.CurrentRemoteHost.DestroySecuritySession(securityProviderName);
		}

		/// <summary>
		/// Unregisters the specified Security Provider.
		/// </summary>
		/// <param name="securityProviderName">The name of the Security Provider.</param>
		public void ReleaseSecurityContext(string securityProviderName)
		{
			// release the SS
			this.ResetSecurityContext(securityProviderName);

			// and remove the security provider
			ITransportContext iTransportContext = GenuineUtility.CurrentRemoteHost.ITransportContext;
			iTransportContext.IKeyStore.SetKey(securityProviderName, null);
		}

		/// <summary>
		/// Returns the SSPI target.
		/// </summary>
		/// <returns></returns>
		public string GetTarget()
		{
			return SspiApi.GetUserNameEx();
		}

		public void MessageHasBeenSent(string message)
		{
			Console.WriteLine("Message \"{0}\" has been sent by {1}.", message, WindowsIdentity.GetCurrent().Name);
		}

		/// <summary>
		/// This is to insure that when created as a Singleton, the first instance never dies,
		/// regardless of the expired time.
		/// </summary>
		/// <returns></returns>
		public override object InitializeLifetimeService()
		{
			return null;
		}

	}
}
